#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to create a sketch effect of an image (using GMIC filters)

# (c) Dinasset 2023
#
#   History:
#
#   v.1.1.0: 2023-09: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, os
import time
import datetime
from gimpfu import *

# define here some constant value, when required
maxwh=8000
fixwh=4000

############################################################################
# Service routines
#---------------------------------------------------------------------------
# from Original drawable create the Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inLayer, widthheight, prepstep):
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inLayer.name
    
   # define/calculate the scale factor for the (new) output image -------------------------
    realWidthHeight = owidth+oheight      # real size of the input image
    if widthheight == 0 :   # means the user wants to have output size same as input size
        if realWidthHeight<=maxwh:
            widthheight = realWidthHeight
        else:
            widthheight = maxwh  # max
            gimp.message ("forced max width+height")
    optiWidthHeight = widthheight       # size to work on to get best results
    scale = (optiWidthHeight*1.0)/(realWidthHeight*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size ------------------------------------------------
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)                    
    newBG = pdb.gimp_layer_new_from_drawable(inLayer, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewImage: Background Layer from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = oname+" (original)"
    
    # create base for effect
    pdb.gimp_image_set_active_layer (newImage,  newBG)

    # optional user choice for preprocessing
    if prepstep>0:  #user did a selection
        #               #"DCP Dehaze (good for light image",
        #               #"Dynamic Range Increase",
        #               #"Equalize Local Histogram",
        #               #"Local Contrast Enhancement",
        #               #"Local Normalization",
        #               #"Local Processing (Equalize)",
        #               #"Local Processing (Normalize)",
        if prepstep==1: # jeje_dehaze 5,1,0.2,1,0,0,0,0,0
            parms='5,1,0.2,1,0,0,0,0,0'
            pdb.plug_in_gmic_qt(newImage,newBG, 1,0,
            "-v - -jeje_dehaze "+str(parms)+" ")
        elif prepstep==2: # fx_tk_dri 0.5,0.5,0.5,2.5,0.5,1,1,0
            parms='0.5,0.5,0.5,2.5,0.5,1,1,0'
            pdb.plug_in_gmic_qt(newImage,newBG, 1,0,
            "-v - -fx_tk_dri "+str(parms)+" ")
        elif prepstep==3: # fx_equalize_local_histograms 75,2,4,100,8,1,16,0,50,50
            parms='75,2,4,100,8,1,16,0,50,50'
            pdb.plug_in_gmic_qt(newImage,newBG, 1,0,
            "-v - -fx_equalize_local_histograms "+str(parms)+" ")
        elif prepstep==4: # fx_LCE 80,0.5,1,1,0,0
            parms='80,0.5,1,1,0,0'
            pdb.plug_in_gmic_qt(newImage,newBG, 1,0,
            "-v - -fx_LCE "+str(parms)+" ")
        elif prepstep==5: # fx_normalize_local 2,6,5,20,1,11,0,50,50
            parms='2,6,5,20,1,11,0,50,50'
            pdb.plug_in_gmic_qt(newImage,newBG, 1,0,
            "-v - -fx_normalize_local "+str(parms)+" ")
        elif prepstep==6: # fx_local_processing 1,75,10,50,20,0,7,0,50,50
            parms='1,75,10,50,20,0,7,0,50,50'
            pdb.plug_in_gmic_qt(newImage,newBG, 1,0,
            "-v - -fx_local_processing "+str(parms)+" ")
        elif prepstep==7: # fx_local_processing 0,75,10,50,20,0,7,0,50,50
            parms='0,75,10,50,20,0,7,0,50,50'
            pdb.plug_in_gmic_qt(newImage,newBG, 1,0,
            "-v - -fx_local_processing "+str(parms)+" ")
    newBG = pdb.gimp_image_get_active_layer (newImage)
            
    # adjust mean value of the base
    # "normal" mean value should be betweeb 117.5 and 137.5
    (mean,B,C,D,E,F) = pdb.gimp_drawable_histogram (newBG, HISTOGRAM_LUMINANCE, 0.0, 1.0)
    #gimp.message ("mean="+str(mean))
    if mean<117.5:
        #gimp.message ("mean before="+str(mean))
        loop=0
        while (mean<117.5) and (loop<100):  
            pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      1.1, 0.0, 1.0, False)
            (mean,B,C,D,E,F) = pdb.gimp_drawable_histogram (newBG,
                                                 HISTOGRAM_VALUE, 0.0, 1.0)
            loop+=1
        #gimp.message ("mean after="+str(mean))
        newBG.name = newBG.name+" mean adj"
    elif mean>137.5:
        #gimp.message ("mean before="+str(mean))
        loop=0
        while (mean>137.5) and (loop<100):  
            pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      0.9, 0.0, 1.0, False)
            (mean,B,C,D,E,F) = pdb.gimp_drawable_histogram (newBG,
                                                 HISTOGRAM_VALUE, 0.0, 1.0)
            loop+=1
        #gimp.message ("mean after="+str(mean))
        newBG.name = newBG.name+" mean adj"
    newBG.opacity = 100.0
    
    return(newImage, widthheight)
#--------------------------------------------------------------------------- 
    

############################################################################
# Main
#---------------------------------------------------------------------------

def SimpleMosaic_33 (inImage, inLayer,
                    widthheight, tilesize, tilereg, thick,
                    prepstep,
                    marblepattern, posteffect,
                    doFlatten):
    
    
    # here the logic of the filter
    oname = inLayer.name
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    
    (newImage, WH) = createOutimage (inImage, inLayer, widthheight, prepstep)
    newBG = newImage.layers[0]
    newBG.name = "newBG Preprocessed"
    original = newImage.layers[1]
    
    # create base for effect
    pdb.gimp_image_set_active_layer (newImage,  newBG)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # apply Superpixels
    #[G'MIC] fx_superpixels 24,24,5,1,1,0,0,0,255,0,50,50
    SuperPix = newBG.copy()
    newImage.add_layer (SuperPix, 0)
    tile=max(16,int(WH/tilesize))
    regu=float(tilereg*1.28)
    #thik=int(round(WH*1.0/1000))
    if thick==0: # auto
        thick=max(1,int(round(tile*1.0/10)))
    else:
        thick=int(thick)
    #gimp.message ("tile size fixed="+str(tile))
    #gimp.message ("tiles sep thickness="+str(thick))
    #pdb.gimp_display_new (newImage)
    #return
    pdb.plug_in_gmic_qt(newImage, SuperPix, 1,0,
        "-v - -fx_superpixels "+str(tile)+","+str(regu)+
        ",5,1,1,"+str(thick)+",0,0,0,255,0,50,50 ")
    #fx_superpixels 16,10,5,1,1,3,0,0,0,255,0,50,50 #g'mic 3.3
    SuperPix.name = "Superpixels Layer"
    SuperPix.opacity = 100.0
    SuperPix.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # apply Isophotes
    #[G'MIC] fx_isophotes 8,0,1,0,50,50
    
    Isophotes = SuperPix.copy()
    newImage.add_layer (Isophotes, 0)
    pdb.plug_in_gmic_qt(newImage, Isophotes, 1,0,
        "-v - -fx_isophotes 8,0,1,0,50,50 ")
    Isophotes.name = "Isophotes Layer"
    Isophotes.opacity = 100.0
    Isophotes.mode = LAYER_MODE_NORMAL
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # apply EmbossRelief
    #[G'MIC] fx_emboss_relief 5,180,0.5,2,0,0,1,0,50,50
    
    EmbossRelief = Isophotes.copy()
    newImage.add_layer (EmbossRelief, 0)
    pdb.gimp_image_select_color (newImage, CHANNEL_OP_REPLACE,
                                 EmbossRelief, (0.0,0.0,0.0,1.0))
    pdb.plug_in_gmic_qt(newImage, EmbossRelief, 1,0,
        "-v - -fx_emboss_relief 5,180,0.5,2,0,0,1,0,50,50 ")
    EmbossRelief.name = "EmbossRelief Layer"
    EmbossRelief.opacity = 75.0
    EmbossRelief.mode = LAYER_MODE_PIN_LIGHT
    pdb.gimp_selection_none (newImage)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # apply marble pattern
    pdb.gimp_context_set_pattern(marblepattern)  # set the pattern
    #gimp.message ("chosen pattern = "+str(marblepattern))
    # --- open the pattern as image
    (Pwidth,Pheight,Pbpp) = pdb.gimp_pattern_get_info (marblepattern)
    Pimage = pdb.gimp_image_new (Pwidth, Pheight, RGB)
    Player = pdb.gimp_layer_new (Pimage, Pwidth, Pheight, RGBA_IMAGE,
                                 marblepattern, 100.0, LAYER_MODE_NORMAL)
    Pimage.add_layer (Player,0)
    pdb.gimp_edit_fill (Player, FILL_PATTERN)
    pdb.gimp_image_scale (Pimage, newImage.width, newImage.height)
    
    Marble = pdb.gimp_layer_new_from_drawable (Player, newImage)
    newImage.add_layer (Marble, 0)
    Marble.mode = LAYER_MODE_VIVID_LIGHT    
    Marble.opacity = 35.0
    
    #use the pattern image to get the mean RGB value of that pattern
    Pmean = Player.copy()
    pdb.plug_in_gmic_qt(Pimage, Pmean, 1,0,
        "-v - -fx_rep_bchstatfunc_average 0,0,0,50,50 ")
    (MR,MG,MB,MA) = pdb.gimp_image_pick_color (Pimage, Pmean, 5.0, 5,0,
                                            True, 5.0)
    MRL=min(255,MR*2); MGL=min(255,MG*2); MBL=min(255,MB*2);MAL=255
    pdb.gimp_context_set_foreground ((MRL,MGL,MBL,MAL))
    
    # BG on Top for details
    BGonTop = newBG.copy()
    newImage.add_layer (BGonTop, 0)
    BGonTop.mode = LAYER_MODE_GRAIN_MERGE
    BGonTop.opacity = 35.0
    
    # post processing effect
    PostP = pdb.gimp_layer_new_from_visible (newImage, newImage,
                               "Layer for Post Processing effect")
    newImage.add_layer (PostP, 0)
    PostP.mode = LAYER_MODE_NORMAL
    PostP.opacity = 0.0 # default
    #gimp.message ("PP="+str(posteffect))
    if posteffect>0: # not *none
        if posteffect==1:   #Textured Glass
            # remove previousy extreme values
            pdb.gimp_drawable_levels (PostP, HISTOGRAM_VALUE,
                                  0.0, 1.0, False,
                                  1.0, 0.2, 0.8, False)
            pdb.plug_in_gmic_qt(newImage, PostP, 1,0,
            "-v - -fx_textured_glass 40,40,1,1,0,2,2,0,50,50 ")
            # then increase local contrast
            pdb.plug_in_gmic_qt(newImage,PostP, 1,0,
                "-v - -fx_LCE 80,0.5,1,1,0,0 ")
            PostP.name = "Glass overlay"
            PostP.mode = LAYER_MODE_LUMA_LIGHTEN_ONLY
            PostP.opacity = 50.0
        if posteffect==2:   # Light Glow
            pdb.plug_in_gmic_qt(newImage, PostP, 1,0,
            "-v - -fx_lightglow 50,0.5,8,0.8,0,0,50,50 ")
            PostP.name = "Light Glow overlay"
            PostP.mode = LAYER_MODE_LUMA_LIGHTEN_ONLY
            PostP.opacity = 70.0
        if posteffect==3:   # Light Leaks
            pdb.plug_in_gmic_qt(newImage, PostP, 1,0,
            "-v - -fx_light_leaks 25,-35,1,1,0,0.5,2,0,0,50,50 ")
            PostP.name = "Light Leaks overlay"
            PostP.mode = LAYER_MODE_LUMA_LIGHTEN_ONLY
            PostP.opacity = 50.0
        if posteffect==4:   # Light Patch
            pdb.plug_in_gmic_qt(newImage, PostP, 1,0,
            "-v - -fx_light_patch 5,0.7,2.5,0 ")
            PostP.name = "Light Patch overlay"
            PostP.mode = LAYER_MODE_LUMA_LIGHTEN_ONLY
            PostP.opacity = 50.0
        if posteffect==5:   # Light Rays
            lightcolor = str(MRL)+","+str(MGL)+","+str(MBL)
            pdb.plug_in_gmic_qt(newImage, PostP, 1,0,
           #"-v - -fx_lightrays 10,15,-15,1,0.5,0,255,255,127 ")
            "-v - -fx_lightrays 10,15,-15,1,0.5,0,"+str(lightcolor)+" ")
            #fx_lightrays 30,50,50,1,0.5,0,255,255,127 for G'MIC 3.3
            PostP.name = "Light Rays overlay"
            PostP.mode = LAYER_MODE_DODGE
            PostP.opacity = 50.0
        if posteffect==6:   # Hair Locks
            pdb.plug_in_gmic_qt(newImage, PostP, 1,0,
            "-v - -gtutor_hairlock 0.9,0.9,0.75,255,255,0,255,0.25,45,0.5,0.5,0 ")
            PostP.name = "Hair Locks overlay"
            PostP.mode = LAYER_MODE_GRAIN_MERGE
            PostP.opacity =70.0
    # final step 
    if doFlatten==True:
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        outcome.name = "Outcome" 
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
    # remove black values 0.0-25.5 and white values 229.5-255.0
    pdb.gimp_drawable_levels (outcome, HISTOGRAM_VALUE, 0.0, 1.0, False,
                              1.00, 0.1, 0.9, False)
        
    outcome.name = st+oname+" with a simple mosaic effect " 
    pdb.gimp_display_new (newImage)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="dinasset"
menu="<Image>/Diego/SimpleFilters/"
copyrightYear="2023"
desc="SimpleMosaic"

register(
    "SimpleMosaic_33",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    "Simple Mosaic, version 1.1.1 (req GMIC 3.3)",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"+
         "\n (0=same as input)", 0, (0, 8000, 100)),
        (PF_ADJUSTMENT, "tilesize", "Mosaic tile relative size"+
         "\n (value as divisor of image size)", 65, (25, 200, 1)),
        (PF_ADJUSTMENT, "regularity", "Mosaic tile regularity"+
         "\n (value as multiplier of 1.28)", 5, (0, 100, 1)),
        (PF_ADJUSTMENT, "tilesep", "Mosaic tile border thickness"+
         "\n (0=auto)", 0, (0, 10, 1)),
        (PF_OPTION, "preproc", "Choose preprocessing step ", 0,
                     ["*none",
                      "DCP Dehaze (for light images)",
                      "Dynamic Range Increase",
                      "Equalize Local Histogram",
                      "Local Contrast Enhancement",
                      "Local Normalization",
                      "Local Processing (Equalize)",
                      "Local Processing (Normalize)",
                      ]),
        (PF_PATTERN, "marble", "Choose marble pattern", "GoSS1000Q_VanGogh.jpg"),
        (PF_OPTION, "postproc", "Choose postprocessing effect ", 0,
                     ["*none",
                      "Textured Glass",
                      "Light Glow",
                      "Light Leaks",
                      "Light Patch",
                      "Light Rays",
                      "Hair Locks ",
                      ]),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    SimpleMosaic_33,
    menu=menu
)

main()
